/*
 * stream queuing
 *  by k-chinen@is.aist-nara.ac.jp, 2001
 *
 * $Id$
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>


#ifndef RETSIGTYPE
#define RETSIGTYPE void
#endif

void SetSignalHandler(int signo, RETSIGTYPE (*handler)(int))
{
    struct sigaction act, oact;

    act.sa_handler = handler;
    sigemptyset(&act.sa_mask);

    act.sa_flags = 0;
    if(signo == SIGALRM) {
#ifdef SV_INTERRUPT
        act.sa_flags |= SV_INTERRUPT;
#endif
    }
    else {
#ifdef SA_RESTART
        act.sa_flags |= SA_RESTART;
#endif
    }

    sigaction(signo, &act, &oact);
}



char *version = "version 1.33";

int verbose     = 1;
int debug       = 0;
int quiet       = 1;
int numbering   = 0;
int timestamping = 0;
int attchar   = '\0';
char *attstr  = NULL;
char *lines = NULL;;
int linewidth = 1024;
int linenum = 2048;
int linetotalsize = -1;

int num = 0;
int lpos = 0;
char* cpos = NULL;
char *filename = NULL;

int report = -1;

int maxlen = -1;

void
init(void)
{
    linetotalsize = linewidth*linenum;
    if(verbose) {
        fprintf(stderr, "alloc %d bytes (%dKB+%d)\n",
            linetotalsize, linetotalsize/1024, linetotalsize%1024);
    }

    lines = (char*)malloc(linetotalsize);
    if(lines==NULL) {
        fprintf(stderr, "not enough memory (%d bytes)\n", linetotalsize);
        exit(1);
    }
}

void
loop(void)
{
    char *cchk;
    register char *p, *q;
    int eol;
    int len;
    
    do {
        lpos = num % linenum;
        cpos = &lines[lpos*linewidth];

        cchk = fgets(cpos, linewidth, stdin);
        if(cchk==NULL) {
            break;
        }
        if(debug) {
            fprintf(stderr, " %4d < %s", num, cpos);
        }
        if(report>0 && ( num % report == report -1 ) ) {
            fprintf(stderr, "read %d lines\n", num+1);
        }

        p = cpos;
        q = NULL;
        eol = 0;
        len = 0;
        while(*p) {
            q = p;
            p++;
            len++;
        }
        if(q) {
            while(*q && q>=cpos && (*q=='\r'||*q=='\n')) {
                *q = '\0';
                q--;
            }
            eol = 1;
        }
        if(len>maxlen) {
            maxlen = len;
        }

        if(quiet) {
            if(eol) {
                fprintf(stdout, "%s\n", cpos);
            }
            else {
                fprintf(stdout, "%s", cpos);
            }
            /* in tee-mode, I/O occures frequentry, flush is not required */
        }

        if(attchar && (*cpos == attchar)
                && strncmp(cpos, attstr, strlen(attstr))==0) {
            if(eol) {
                fprintf(stdout, "%s\n", cpos);
            }
            else {
                fprintf(stdout, "%s", cpos);
            }
            fflush(stdout);
        }
        
#if 0
        if(debug) {
            fprintf(stderr, ";%4d: %s\n", num, cpos);
        }
#endif
        
        num++;
    } while(1);

    if(verbose) {
        fprintf(stderr, "read %d lines\n", num);
    }
}

void
writeback(void)
{
    FILE *fp;
    int off;
    int first, last;
    char *pos;
    int i, c;

    if(!filename) {
        return;
    }

    if(verbose) {
        fprintf(stderr, "writeback to '%s'\n", filename);
        fprintf(stderr, "maxlen is %d\n", maxlen);
    }
    
    fp = fopen(filename, "w");
    if(fp==NULL) {
        fprintf(stderr, "cannot open file '%s'\n", filename);
        return;
    }
    if(num>linenum) {
        last = (num-1) % linenum;
        if(last==linenum-1) {
            first = 0;
        }
        else {
            first = last+1;
            last += linenum;
        }
        off = num - last;
    }
    else {
        off   = 0;
        first = 0;
        last  = num-1;
        off = num - last;
    }
    if(debug) {
        fprintf(stderr, "%d/%d [%d..%d]\n", num, linenum, first, last);
    }

    c = 0;
    for(i=first;i<=last;i++) {
        lpos = i % linenum;
        cpos = &lines[lpos*linewidth];

        if(numbering) {
            if(debug) {
                fprintf(stdout, " %4d > %4d %s\n", i, off+i, cpos);
            }
            fprintf(fp, " %4d %s\n", off+i, cpos);
        }
        else {
            if(debug) {
                fprintf(stdout, " %4d > %s\n",  i, cpos);
            }
            fprintf(fp, "%s\n", cpos);
        }
        c++;
    }
    
    fclose(fp);

    if(verbose) {
        fprintf(stderr, "write %d lines to file\n", c);
    }

}

RETSIGTYPE
catch_int(int x)
{
    exit(0);
}

/*
  TODO:
    -t      (toggle) timestamping
*/

void
usage(char *name)
{
    fprintf(stdout, "\
stq - stream queuing program %s\n\
    by k-chinen@is.aist-nara.ac.jp, 2001\n\
\n\
usage:\n\
    %% %s [options] [filename]\n\
\n\
options:\n\
    -h      print this message\n\
    -v      (toggle) verbose (default on)\n\
    -q      (toggle) quiet (default off)\n\
    -d      (toggle) debug\n\
    -n      (toggle) numbering\n\
    -a str  attention charactor - print specified messages\n\
    -w #    the width of lines          (default %d)\n\
    -l #    the number of lines         (default %d)\n\
    -r #    the interval of reporting   (default %d)\n\
\n\
example:\n\
    %% ./kotetuserv -f kotetu.conf | %s -l 50000 last-snap\n\
    %% ./kotetuserv -f kotetu.conf | %s -q -a '  =sw' -r 1000 last-snap\n\
    %% vmstat 5 | %s -v vmstat.log\n\
\n\
",
    version,
    name, 
    linewidth, linenum, report,
    name, name, name);
}


main(int argc, char* argv[])
{
    int flag;

    while((flag=getopt(argc, argv, "hvdntqa:w:l:r:")) != -1) {
        switch(flag) {
        case 'h':
            usage(argv[0]);
            exit(0);
            break;
        case 'v':
            verbose = 1 - verbose;
            break;
        case 'd':
            debug = 1 - debug;
            break;
        case 'n':
            numbering = 1 - numbering;
            break;
        case 't':
            timestamping = 1 - timestamping;
            break;
        case 'q':
            quiet = 1 - quiet;
            break;
        case 'a':
            attstr  = strdup(optarg);
            attchar = attstr[0];
            break;
        case 'w':
            linewidth = atoi(optarg);
            break;
        case 'l':
            linenum = atoi(optarg);
            break;
        case 'r':
            report = atoi(optarg);
            break;
        default:
            usage(argv[0]);
            exit(3);
            break;
        }
    }
    if(optind<argc) {
        filename = strdup(argv[optind]);
    }
    if(debug) {
        verbose = 1;
    }
    if(verbose) {
        fprintf(stderr, "linewidth %d num %d report %d\n",
                linewidth, linenum, report);
        if(filename) {
            fprintf(stderr, "filename '%s'\n", filename);
        }
        else {
            fprintf(stderr, "filename is not specified\n");
        }
        if(attchar) {
            fprintf(stderr, "attention string \"%s\", char '%c'\n",
                attstr, attchar);
        }
    }

    SetSignalHandler(SIGINT, catch_int);    
    SetSignalHandler(SIGTERM, catch_int);   
    SetSignalHandler(SIGQUIT, catch_int);   
    atexit(writeback);

    init();
    loop();

    exit(0);
}
